;******************************************************************************
;   Subroutines to handle initialzation and operation of blitter.
;******************************************************************************
; 100 22 May 95 JPE     Initial revision (incorporates preliminary subtitle
;                       code from Dave Schwartz).
;******************************************************************************

        .nlist
        .include    'memory.inc'
        .include    'jaguar.inc'
        .include    'cinepak.inc'
        .list

;==============================================================================
;	Subroutine to initialize the following blitter variables:
;
;       blitCount
;       blitPixel
;       blitScreen
;       blitStep
;       dest
;
;		Input:	d0 = width of movie (pixels)
;               d1 = height of movie (lines)
;
;		Output:	None.
;==============================================================================

InitBlit:
        movem.l d2-d4,-(a7)             ; Save used registers

        move    #0,blitScreen

        move    d1,d2                   ; Form Blitter B_COUNT
        swap    d2
        move    d0,d2
        move.l  d2,blitCount

        move.l  d0,d2                   ; Form Blitter STEP regs
        neg     d2
        ori.l   #$10000,d2
        move.l  d2,blitStep

        move.l  #NLINES,d2              ; Find true Vert center
        sub.l   d1,d2                   ; d2 = (d2-d1)/2
        lsr.l   #1,d2

        move.l  d2,d3                   ; Copy
        mulu    #(NPIXELS*2),d3         ; Mult * bytes per row

        swap    d2                      ; Put in high word

        move    #NPIXELS,d2             ; Find true Horz center
        sub     d0,d2                   ; d2 = (d2-d0)/2
        lsr     #1,d2

        clr.l   d4
        move    d2,d4                   ; # Pixel Offset
        lsl.l   #1,d4                   ; * 2 (for bytes)

        add.l   d4,d3
        mulu    #3,d3
        add.l   #SCREEN_BASE,d3
        move.l  d3,dest

;        move.l  d2,blitPixel            ; Built Blitter A1/A2_PIXEL
        move.l  #0,blitPixel

        movem.l (a7)+,d2-d4             ; Restore used registers
        rts

;==============================================================================
;	Subroutine to perform blitter operation.
;
;		Input:	None.
;
;		Output:	None.
;==============================================================================

DoBlit:
        movem.l d0-d1,-(a7)             ; Save used registers

        move.l  #SCREEN_BASE+$8,d0      ; Choose screen to blit to
        move.l  #SCREEN_BASE+$10,d1

        btst.b  #0,blitScreen
        beq     BlitMe  

        exg.l   d0,d1                   ; Swap buffers

BlitMe:
        move.l  d0,A1_BASE              ; Do normal phrase screen copy
        move.l  #SCREEN_BASE,A2_BASE
        move.l  #PIXEL16|XADDPHR|PITCH3|WID320,A1_FLAGS
        move.l  #PIXEL16|XADDPHR|PITCH3|WID320,A2_FLAGS
        move.l  blitPixel,A1_PIXEL
        move.l  blitPixel,A2_PIXEL
        move.l  blitStep,A1_STEP
        move.l  blitStep,A2_STEP
        move.l  blitCount,B_COUNT
        move.l  #SRCEN|UPDA1|UPDA2|LFU_REPLACE,B_CMD

Wait4Blit:
        move.l  B_CMD,d1                ; Ensure blit has finished
        btst.l  #0,d1
        beq     Wait4Blit

        move.l  d0,scrbuf               ; Store it for op vframe
        bchg.b  #0,blitScreen           ; Switch screen

        movem.l (a7)+,d0-d1             ; Restore used registers
        rts

        .extern blitCount
        .extern blitPixel
        .extern blitScreen
        .extern blitStep
        .extern dest
        .extern scrbuf

        .globl  InitBlit
        .globl  DoBlit

;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;                               Subtitle Subroutine
;""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
                                        ; NOTE:
                                        ; see line 271 for framenum reset at
                                        ; each movie and line 139 for track number
                                        ; incrementing
                .long
subtiton:       dc.l    0               ; subtitle on-screen flag
framenum:       dc.l    0               ; Store movie's current frame number here
traknumr:       dc.l    0               ; current CD track number
S_frmnum:       dc.l    24              ; start frame for subtitle
E_frmnum:       dc.l    48              ; ending frame for subtitle
exx_post:       dc.w    0               ; x position of start of subtitle in window
why_post:       dc.w    0               ; y position " "
backcolr:       dc.w    0               ; background color
fontcolr:       dc.w    0               ; font color
strnglen:       dc.w    0               ; length of text string in bytes
pan_rate:       dc.w    0               ; rate to scroll text

                .globl  subtiton
                .globl  framenum
                .globl  traknumr
                .globl  S_frmnum
                .globl  E_frmnum
                .globl  exx_post
                .globl  why_post
                .globl  backcolr
                .globl  fontcolr
                .globl  strnglen
                .globl  pan_rate
                .globl  _FNTstr

                .globl  Subtitle
                .globl  get_joy

.phrase
font:           .incbin "LURS14.JFT"    ; font table

.phrase
subtit_0:
dc.b    'Caves of Fear subtitle',0
.even

.phrase
subtit_1:
dc.b     '      cavehead        .',0
.even

.phrase
subtit_2:
dc.b     '      Path 1         ',0
.even

.phrase
subtit_3:
dc.b     '      Path 2         ',0
.even

.phrase
subtit_4:
dc.b     '     Scene 0608      ',0
.even

.phrase
subtit_5:
dc.b     '     Scene 0911      ',0
.even

.phrase
subtit_6:
dc.b     '     Scene 1214      ',0
.even

.phrase
subtit_7:
dc.b     '     Scene 1517      ',0
.even

.phrase
subtit_8:
dc.b     '     Scene 18        ',0
.even

.phrase
subtit_9:
dc.b     '     Scene 1621      ',0
.even

.phrase
subtit10:
dc.b     '     Scene 2324      ',0
.even

.phrase
subtit11:
dc.b     '     Scene 22        ',0
.even

.phrase
subtit12:
dc.b     '     Scene 25        ',0
.even

Subtitle:
                movem.l a0-a5/d0-d5,-(sp)               ; push to stack

whattrak:                                               ; get track number
                move.l  traknumr,d0                     ; match to subtitle set
                cmp.l   #1,d0
                beq     trak1set
                cmp.l   #2,d0
                beq     trak2set
                cmp.l   #3,d0
                beq     trak3set
                cmp.l   #4,d0
                beq     trak4set
                cmp.l   #5,d0
                beq     trak5set
                cmp.l   #6,d0
                beq     trak6set
                cmp.l   #7,d0
                beq     trak7set
                cmp.l   #8,d0
                beq     trak8set
                cmp.l   #9,d0
                beq     trak9set
                cmp.l   #10,d0
                beq     trak10set
                cmp.l   #11,d0
                beq     trak11set
                cmp.l   #12,d0
                beq     trak12set

trak1set:
                move.l  #subtit_1,a1                     ; point to string
                bra     framecnt

trak2set:
                move.l  #subtit_2,a1                     ; point to string
                bra     framecnt
trak3set:
                move.l  #subtit_3,a1                     ; point to string
                bra     framecnt
trak4set:
                move.l  #subtit_4,a1                     ; point to string
                bra     framecnt
trak5set:
                move.l  #subtit_5,a1                     ; point to string
                bra     framecnt
trak6set:
                move.l  #subtit_6,a1                     ; point to string
                bra     framecnt
trak7set:
                move.l  #subtit_7,a1                     ; point to string
                bra     framecnt
trak8set:
                move.l  #subtit_8,a1                     ; point to string
                bra     framecnt
trak9set:
                move.l  #subtit_9,a1                     ; point to string
                bra     framecnt
trak10set:
                move.l  #subtit10,a1                     ; point to string
                bra     framecnt
trak11set:
                move.l  #subtit11,a1                     ; point to string
                bra     framecnt
trak12set:
                move.l  #subtit12,a1                     ; point to string
                bra     framecnt


framecnt:                                               ; count frames
blittime:                                               ; is it time yet?
                move.l  framenum,d0
                cmp.l   #24,d0                         ; wait for frame
                bge     at_start                        ; found start time
                movem.l (sp)+,d0-d5/a0-a5               ; pop stack
                rts                                     ; not there yet, return

at_start:
                move.l  framenum,d0
                cmp.l   #48,d0                         ; end of display time?
                ble     blit_msg                        ; no, jump to display
                                                        ; yes, kill subtitle
                                                        ; by overwriting it

killsubt:                                               ; routine to blank subtitle
                move.l  #0,subtiton                     ; clear flag
                move.l  #SCREEN_BASE+$8,d0              ; get screen to blit to
                move.l  #SCREEN_BASE+$10,d1
                btst.b  #0,blitScreen
                beq     blitme2
                exg.l   d0,d1                           ; Swap buffers

blitme2:                                                ; arguments to _FNTstr
                move.w #$0,-(sp)                        ; bgcolor
                move.w #$0,-(sp)                        ; fgcolor
                move.l #font,-(sp)                      ; FNThead
                move.l #PIXEL16|PITCH3|WID320,-(sp)     ; blitflags
                move.l d0,-(sp)                         ; void *dest
                move.l a1,-(sp)                       ; char *str
                move.w #198,-(sp)                       ; int y
                move.w #80,-(sp)                        ; int x
                jsr     _FNTstr                         ; do it
                lea     $18(a7),a7                      ; Balance stack
killexit:
                movem.l (sp)+,d0-d5/a0-a5               ; pop stack
                rts

blit_msg:                                               ; blit some text
                move.l  #1,subtiton                     ; set flag
                move.l  #SCREEN_BASE+$8,d0              ; get screen to blit to
                move.l  #SCREEN_BASE+$10,d1
                btst.b  #0,blitScreen
                beq     blitme3
                exg.l   d0,d1                           ; Swap buffers

blitme3:                                                ; arguments to _FNTstr
                move.w  #$0,-(sp)                        ; bgcolor
                move.w  #$7DB0,-(sp)                     ; fgcolor
                move.l  #font,-(sp)                      ; FNThead
                move.l  #PIXEL16|PITCH3|WID320,-(sp)     ; blitflags
                move.l  d0,-(sp)                         ; void *dest
                move.l  a1,-(sp)                        ; char *str
                move.w  #198,-(sp)                       ; int y
                move.w  #80,-(sp)                        ; int x
                jsr     _FNTstr                         ; do it
                lea     $18(a7),a7                      ; Balance stack
                movem.l (sp)+,d0-d5/a0-a5               ; pop stack
                rts

;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;                               joystix
;""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
get_joy:                                ; get joystick data and interpret it
        movem.l a0-a5/d0-d5,-(sp)       ; push to stack
        clr.l   d6              ; use d6
        move.l  #$f14000,a0     ; address joystick
        move.w  #$8000,(a0)     ; write before read
        move.w  (a0),d1         ; new joystick data now in d1
        andi.w  #$0F00,d1       ; mask for our nibble
        cmp.w   #$F00,d1        ; is the port idle, ie our nibble is F?
        beq    joy_over         ; yes, port is idle so return w/ d6=0
        move.l  #1,d6           ; joystick active

joy_over:                       ; return with 0 or 1 in d6
        movem.l (sp)+,d0-d5/a0-a5               ; pop stack
        rts

        .end
