;******************************************************************************
; (C) Copyright 1992-1994, SuperMac Technology, Inc.
; All rights reserved.
;
; This source code and any compilation or derivative thereof is the sole
; property of SuperMac Technology, Inc. and is provided pursuant to a
; Software License Agreement.  This code is the proprietary information
; of SuperMac Technology and is confidential in nature.  Its use and
; dissemination by any party other than SuperMac Technology are strictly
; limited by the confidential information provisions of the Agreement
; referenced above.
;
; Revision 1.3  10/19/94  18:16:10  jpe
; Disabled CD setup and mode selection for RAM playback.
; Fixed bug in calculation of fractional portion of timeIncr.
; Truncated GPU_READY to GPU_READ to keep COFF linker happy.
;
; Revision 1.2  04/27/94  18:46:55  jpe
; CD-BIOS changes incorporated.
;
; Revision 1.1  04/25/94  17:07:10  jpe
; Corrected aspect-ratio problem with video timing.
; Fixed bug which prevented operation without CD-ROM.
; Fixed bug which caused improper handling of empty chunk.
; Added initialization of audio drift rate.
;
; Revision 1.0  04/08/94  13:57:05  jpe
; Initial revision.
;******************************************************************************
; Atari In-House Revisions:
; 1.3.1: SDS
;        Converted to Cinepak Demo Code...not for distribution.
; 1.3.2: SDS
;        Made DSP Code play different format audio.
; 1.3.3: SDS
;        Hold on to your seats...rotation!
; 1.3.4: SDS
;        GPU/DSP now never stop.
; 1.4.0: SDS
;        Triple-Interleaved Buffering.
; 1.4.1: SDS
;        DSP Code Fix/New & Better Startup
; 1.4.2: SDS
;        Correct Expanded RGB Handling
; 1.5.0: SDS
;        Replaced old player, working backwards from CPKDEMO.
;        Now uses our startup, double buffers, enhanced audio,
;        and uses 600Hz timer rather than 60 or 70Hz screen timer.
; 1.5.1: SDS
;        Fixes two problems:
;          1. Movies with only one chunk should now play.
;          2. Their is no search for the first chunk since it should be
;             contiguous in the buffer with the chunk table (The old
;             problem was that the first chunk was searched for from the
;             start of the buffer, not the end of the chunk table.)
;
; 200 05 May 95 DMS     Make it play 1 track on CD, then loop.  Works on 1 trk
;                       on disc 05may95 dms
; 201 06 May 95 DMS     Make it play all tracks on CD, then loop. Use 08may95a
; 202 10 May 95 JPE     Now works with movies which are short enough to fit 
;                       inside the buffer. (Only one CD access is required!)
; 203 16 May 95 JPE     First halting steps toward separating game code, CD
;                       access code and Cinepak player code.
; 204 17 May 95 JPE     Resurrected snapshot in ROMulator memory. CD access
;                       code now separated from Cinepak player code.
; 205 22 May 95 JPE     Split out blitter initialization and operation code
;                       to separate blit.s module. Add call to subtitle and
;                       joystick routines.
; 206 23 May 95 JPE     Fixed bugs in initialization and incrementing of
;                       track number.
; 207 24 May 95 JPE     Eliminated relocation of chunk table. Added code to
;                       deal with optional subtitle atom.
;******************************************************************************

        .nlist
        .include    'memory.inc'
        .include    'jaguar.inc'
        .include    'cd.inc'
        .include    'player.inc'
        .list

START_SESS      .equ    1
START_TRACK     .equ    1

_start:

;;; At this point an single 320x240 16-bit bitmap object object list has
;;; been setup by STARTUP.S. The address is held in D0.

        move.l  d0,-(sp)                ; Save on stack

        bsr     Clear                   ; Clear display buffers

        jsr     CD_setup                ; Initialize CD-ROM system

        move    #$3,d0                  ; CD-ROM, double-speed
        jsr     CD_mode                 ; Set CD-ROM speed

        bsr     LoadGPU
        bsr     LoadDSP                 ; Download the DSP code

; Start the DSP.

        move.l  #DRIFT_RATE,AUDIO_DRIFT ; (Default)
        move.l  #0,AUDIO_DESC
        move.l  #DSP_ENTR,D_PC          ; Load DSP program counter
        move.l  #$1,D_CTRL              ; Start the DSP

; Start the GPU.

        move.l  (sp)+,GPUOLP            ; Give the GPU the OLP.

        lea     GPU_READ,a0             ; GPU_READY flag 
        adda.l  GPUOffset,a0            ; Relocated address
        move.l  #0,(a0)                 ; Clear the ready flag
        move.l  #G_RAM+$50,G_PC         ; Load GPU program counter
        move.l  #RISCGO,G_CTRL          ; Start the GPU

; Wait for GPU to finish startup sequence, then allow external interrupts.

WaitGPU:
        move.l  (a0),d0                 ; Get ready flag
        tst.l   d0
        beq.s   WaitGPU                 ; Loop until GPU has set it

        bsr     IntInit

        move.b  #START_SESS,d0
        move.b  #START_TRACK,d1

        jsr     MediaGetFirst           ; Get Track as Defined above

        tst.l   d0
        bpl     gotfirst

        move.l  #$00CD00CD,d0           ; NON-EXISTENT TRACK
        illegal

gotfirst:
        add.l   #LEADER-MARGIN,d0       ; Cinepak Leader/Safety Margin
        move.l  d0,blockOffset
        move.l  #START_TRACK,traknumr   ; Initialize track counter

nextmovie:
        clr.l   d0
        move.l  d0,mediaOffset          ; Clear mediaOffset
        move.l  d0,framenum             ; Clear frame counter
        move.l  d0,subt_begin           ; Clear subtitle atom pointer
        move    d0,CRYmovie             ; Clear CRYmovie
        move    d0,playPhase            ; Clear playPhase
        move    d0,catchUp              ; Clear catchUp

        bsr     InitROM                 ; Clear ROMulator memory

        move.l  #FILM_BASE,a0           ; Where to put CD-ROM data
        movea.l a0,a1                   ; Make a copy
        adda.l  #CBUF_SIZE,a1           ; Length of search window
        moveq   #0,d0                   ; Byte offset on CD-ROM
        bsr     ReadCDData

        move.l  #FILM_SYNC,d0           ; Pattern to look for
        movea.l #FILM_BASE,a0           ; Where to start looking
        movea.l a0,a1                   ; Make a copy
        adda.l  #SRCH_WIN,a1            ; Length of search window

.ClearWindow:
        bsr     GetCDWritePtr

        cmpa.l  CDWritePtr,a1           ; See where GPU write pointer is
        bge.s   .ClearWindow            ; Wait for GPU to clear search window

        bsr     FindSync                ; Find sync pattern

        cmpa.l  #0,a0                   ; Did sync search fail?
        bne.s   CheckSubT               ; No, look for subtitle atom tag

        move.l  #SYNC_FAIL,d0           ; Load error flag
        jmp     skipmovie

CheckSubT:
        cmpi.l  #SUBT_TAG,(a0)          ; Is there a subtitle atom?
        bne.s   CheckFilm               ; No, look for film header

        move.l  a0,subt_begin           ; Yes, save starting address
        adda.l  4(a0),a0                ; Skip over subtitle atom

FillSubT:
        bsr     GetCDWritePtr           ; See where write pointer is
        cmpa.l  CDWritePtr,a0           ; Are we ahead of it?
        bge.s   FillSubT                ; Yes, wait

CheckFilm:
        cmpi.l  #FILM_TAG,(a0)          ; Have we found start of film?
        beq.s   SetupMovie              ; Yes, setup movie parameters

        move.l  #BOGUS_FILM,d0          ; Load error flag
        jmp     skipmovie

SetupMovie:
        movea.l a0,a5                   ; Film header
        move.l  4(a5),d0                ; Header size
        addi.l  #SYNC_SIZE,d0           ; Account for film sync
        move.l  d0,mediaOffset          ; Fix offset on CD-ROM

        move.l  a5,d0                   ; Movie base
        add.l   4(a5),d0                ; Add header/CT size
        move.l  d0,cBufBase             ; Base of circular buffer

        move.l  #CBUF_END,d1            ; End of circular buffer
        sub.l   d0,d1                   ; Free memory
        move.l  d1,cBufSize             ; Size of circular buffer

FillHeader:
        bsr     GetCDWritePtr           ; See where write pointer is
        cmp.l   CDWritePtr,d0           ; Beyond cBufBase?
        bge.s   FillHeader              ; No, wait

        lea     $10(a5),a5              ; Frame description
        move.l  $8(a5),d0               ; cType
        cmpi.l  #'$CRY',d0              ; Is it a CRY movie?
        bne     RGBmovie                ; No, continue

        move    #1,CRYmovie             ; Set CRYmovie flag true
        move    #$6C1,VMODE             ; Turn on the display in CRY
        bra.s   SetupBlit

RGBmovie:
        move    #$6C7,VMODE
        move    #0,CRYmovie

        cmpi.l  #'$RGB',d0
        bne.s   SetupBlit

        move    #1,CRYmovie             ; Already expanded

; Read the movie width and height fields from the frame description and
; set up variables used by the blitter.

SetupBlit:
        move.l  $10(a5),d0              ; Width of Movie
        move.l  $C(a5),d1               ; Height of Movie
        bsr     InitBlit                ; Initialize blitter

        move.l  #0,AUDIO_DESC           ; Default Audio
        move.l  #18,SCLK                ; SCLK = 21.793 kHz
        move.l  #DRIFT_RATE,AUDIO_DRIFT ; 0.017606439 * (65536)^2

        lea     $14(a5),a5              ; Chunk table or Audio Desc?
        cmpi.l  #ADSC_TAG,(a5)          ; Is it audio?
        bne.s   ischunk

        move.l  $8(a5),AUDIO_DESC
        move.l  $C(a5),SCLK
        move.l  $10(a5),AUDIO_DRIFT
        lea     $14(a5),a5              ; Skip over ADSC field

ischunk:
        move.l  $c(a5),filmChunks       ; Number of chunks in film

        cmp.l   #1,filmChunks           ; Is there more than one chunk?
        bgt     morethanone

        move.w  #2,playPhase            ; No, straight to playPhase #2

morethanone:
        move.l  $8(a5),d0               ; Time scale
        divu    #TICK_RATE,d0           ; Time per tick (integer part)
        move.l  d0,d1                   ; Remainder << 16 to d1
        clr     d1                      ; Nuke integer part
        divu    #TICK_RATE,d1           ; Fractional part
        swap    d0                      ; Integer part to high word
        move    d1,d0                   ; Merge fractional part (Q16)
        move.l  d0,timeIncr   

        move.l  $8(a5),d0               ; Time scale
        mulu    #MAX_DELAY,d0           ; Delay in units of time (Q16)
        addi.l  #$8000,d0               ; Round off
        swap    d0                      ; Keep integer portion
        ext.l   d0                      ; Convert to long
        move.l  d0,maxDelay

        move.l  cBufSize,d1
        divu    #BLK_SIZE,d1            ; Blocks in circular buffer
        swap    d1                      ; Make Q16
        move    #2*BLK_RATE,d0          ; Blocks per second
        bsr     LongDivide              ; Fill time in Q16 format

        subi.l  #AUDIO_LAG,d1           ; fillTime - audioLag
        lsr.l   #3,d1                   ; Make Q13 (difference < 8!!)
        mulu    $a(a5),d1               ; Convert to time
        lsl.l   #2,d1                   ; Divide by 2, make Q16
        addi.l  #$8000,d1               ; Round off
        swap    d1                      ; Keep integer portion
        ext.l   d1                      ; Convert to long
        move.l  d1,deltaTime

        lea     $10(a5),a5              ; First entry in chunk table
        movea.l cBufBase,a4             ; Start of data in first chunk
        lea     SYNC_SIZE(a4),a3        ; Skip over sync field

        movea.l a4,a0                   ; First chunk in circular buffer
        move.l  a5,pNextGroup           ; Phony startup condition
        bsr     SetNextGroup            ; Set params for next group of chunks

        movea.l a4,a0                   ; Start of data in buffer
        adda.l  #HEAD_START,a0          ; Force space between R/W pointers

        tst.l   filmChunks              ; Remaining chunks in film
        bne.s   FillData                ; Continue if non-zero
        move    #3,playPhase            ; Inhibit further CD access

FillData:
        bsr     GetCDWritePtr
        cmpa.l  CDWritePtr,a0           ; See where GPU write pointer is
        bge.s   FillData                ; Allow GPU to get ahead

        move.l  buffChunks,d7           ; Number of chunks in circular buffer
        move.l  time,d0                 ; Read time

WaitForTick:
        cmp.l   time,d0                 ; Still the same?
        beq.s   WaitForTick             ; Yes, wait for it to change

        clr.l   time                    ; Start at zero
        clr     time+4

BufferLoop:
        bra     PlayBuffer

ResetBuffer:
        cmpi    #3,playPhase            ; Very short film?
        beq.s   skipmovie               ; Yes, we're done!
        move.l  buffChunks,d7           ; Number of chunks in circular buffer
        beq.s   skipmovie               ; If zero, buff done!!

        lea     $10(a5),a5              ; Advance chunk table pointer
        move.l  nextBufAddr,a4          ; Start of fresh data
        lea     $40(a4),a3              ; Skip around sync
        clr     playPhase               ; Reset to phase 0
        tst.l   filmChunks              ; Still more chunks after these?
        bne     BufferLoop              ; Yes, continue

        move    #3,playPhase            ; Inhibit further CD-ROM activity
        clr.l   buffChunks              ; No more chunks after these
        bra     BufferLoop              ; Finish what's in buffer

skipmovie:                              ; skip non-existant trax
        jsr     CD_uread
        jsr     MediaGetNext

debug1:
        tst.l   d0
        bpl     goodtrack

diskbeg:                                ; Not good track, start again
        move.b  #START_SESS,d0
        move.b  #START_TRACK,d1
        move.l  #START_TRACK-1,traknumr ; Re-initialize track counter
        jsr     MediaGetFirst
        tst.l   d0
        bpl     goodtrack
        bra     done                    ; err

goodtrack:
        add.l   #LEADER-MARGIN,d0
        move.l  d0,blockOffset
        bsr     Clear
        addq.l  #1,traknumr
        jmp     nextmovie

done:
        move.l  #$CAFECAFE,d6           ; end code
        jsr     CD_uread
        jsr     CD_stop
        jsr     HaltCpk
        move.l  #0,DSP_STOP
        move.l  #0,d0
        illegal

;******************************************************************************

PlayBuffer:
ChunkLoop:
        move.l  $c(a3),d5               ; Number of samples
        move.l  a3,d4
        add.l   4(a3),d4                ; Base of sample data
        lea     $10(a3),a3              ; Base of sample table

SampleLoop:
        bsr     CheckCDPlay             ; See how CD is doing
        bsr     Snapshot                ; Take a snapshot for debug

        cmpi.l  #-1,8(a3)               ; s->Time == -1?
        bne.s   DoVideo                 ; No, it's video

        lea     DSP_ARGS,a0             ; DSP argument list
        move.l  AudioDesc,8(a0)         ; Audio type
        move.l  d4,4(a0)                ; Transfer *audioData (Do first!!)
        move.l  4(a3),(a0)              ; Transfer count
        bra     NextSample              ; We're done with audio

DoVideo:
        move.l  8(a3),d2                ; s->Time
        bclr    #31,d2                  ; Get rid of MSB

KillTime:
        move.l  time,d0                 ; Get current time
        cmp.l   d0,d2                   ; Ahead of schedule?
        bgt.s   KillTime                ; Yes, wait

        tst.w   catchUp                 ; Is catch-up flag set?
        bne.s   LookForKey              ; Yes, look for key frame

        sub.l   d2,d0                   ; Amount we've slipped
        cmp.l   maxDelay,d0             ; Are we too far behind?
        bmi.s   DisplayFrame            ; Not yet, continue

        move.w  #1,catchUp              ; Yes, set catch-up flag

LookForKey:
        move.l  #0,-(a7)                ; Return value
        move.l  d4,-(a7)                ; Pointer to frame
        bsr     CheckKeyFrame

        lea     $4(a7),a7               ; Balance stack
        move.l  (a7)+,d0                ; Pop return value
        tst.l   d0                      ; Is it a key frame?
        beq     NextSample              ; No, skip the frame

        clr     catchUp                 ; Yes, clear flag and display frame

DisplayFrame:
        move.l  #1,-(a7)                ; Return value      
        move.l  #CINEPAK_DATA,-(a7)     ; c
        move.l  d4,-(a7)                ; (Ptr)data
        move    CRYmovie,-(a7)          ; CRYmovie
        bsr     PreDecompress

        lea     $a(a7),a7               ; Balance stack
        move.l  (a7)+,d0                ; Pop return value
        tst.l   d0                      ; Did we get an error?
        beq.s   StartDecomp             ; No, continue  

        move.l  #PREDEC_ERR,d0          ; Load error flag
        bra     NextSample              ; HANDLE THIS ERROR

StartDecomp:
        move.l  #1,-(a7)                ; Return value      
        move.l  #CINEPAK_DATA,-(a7)     ; c
        move.l  d4,-(a7)                ; (Ptr)data
        move.l  dest,-(a7)              ; data offset

        move.w  #NPIXELS*2,-(a7)        ; WORLD->bytewidth
        move.w  #2,-(a7)                ; Triple Bufferring

        bsr     Decompress

        lea     $10(a7),a7              ; Balance stack
        move.l  (a7)+,d0                ; Pop return value

        ;;; This error is ignored, though, if the content of the
        ;;; video is crucial you may want to replay portions
        ;;; if this error happens.

        bsr     DoBlit                  ; Blit to active buffer

        addq.l  #1,framenum             ; Increment frame counter

        bsr     Subtitle                ; Put up subtitle

in_time:                                ; timely keypress?
        move.l  framenum,d6
        cmp.l   E_frmnum,d6
        bge     NextSample              ; too late
        cmp.l   S_frmnum,d6
        ble     NextSample              ; too early

chekplay:                               ; in window, see if keypress
        bsr     get_joy                 ; get button
        cmp.l   #1,d6
        bne     NextSample
        bra     skipmovie               ; OK! skip to next movie

NextSample:
        add.l   4(a3),d4                ; Data for sample[m+1]
        lea     $10(a3),a3              ; Advance sample table pointer
        subq.l  #1,d5                   ; Decrement sample counter
        bne     SampleLoop

NextChunk:
        subq.l  #1,d7                   ; Decrement buffer chunk counter
        beq     ResetBuffer             ; If zero, reset buffer pointers

        adda.l  4(a5),a4                ; Next chunk in FIFO
        lea     $10(a5),a5              ; Advance chunk table pointer
        lea     $40(a4),a3              ; Skip around sync
        tst.l   $c(a3)                  ; Is the chunk empty?
        beq.s   NextChunk               ; Yes, skip it

        bra     ChunkLoop               ; No, do it

;******************************************************************************

CheckCDPlay:
        cmpi    #0,playPhase            ; Phase 0?
        bne.s   Check1                  ; No, check phase 1

        movea.l pNextGroup,a0           ; First chunk in next group
        move.l  $8(a0),d0               ; Get expiration time
        sub.l   time,d0                 ; Subtract current time
        cmp.l   deltaTime,d0            ; Time to play CD?
        bgt     CheckCDExit             ; No, proceed

        move.l  (a0),d0                 ; Byte offset on CD-ROM
        move.l  cBufBase,a0             ; Where to put CD-ROM data
        bsr     ReadCDData     

        move    #1,playPhase            ; Phase 1 is next
        bra.s   CheckCDExit

Check1:
        cmpi    #1,playPhase            ; Phase 1?
        bne.s   CheckCDExit             ; No, proceed

        movea.l cBufBase,a0             ; Where to start looking
        movea.l a0,a1                   ; Make a copy
        adda.l  #SRCH_WIN,a1            ; Length of search window

        bsr     GetCDWritePtr
        cmpa.l  CDWritePtr,a1           ; See where GPU write pointer is
        bge.s   CheckCDExit             ; Wait for GPU to clear search window

        movea.l pNextGroup,a1           ; Pointer to next group of chunks
        move.l  $c(a1),d0               ; Pattern to look for
        bsr     FindSync                ; Find sync pattern

        cmpa.l  #0,a0                   ; Did sync search fail?
        bne.s   .CheckSample            ; No, look for sample tag

        move.l  #SYNC_FAIL,d0           ; Load error flag
        jmp     skipmovie

.CheckSample:
        cmpi.l  #SAMP_TAG,(a0)          ; Have we found start of chunk?
        beq.s   .ChunkOK                ; Yes, get ready to do it       

        move.l  #BOGUS_SAMP,d0          ; Load error flag
        jmp     skipmovie               ; and try another movie

.ChunkOK:
        lea     -$40(a0),a0             ; Back up to start of sync
        move.l  a0,nextBufAddr          ; Save for when we start next group
        bsr     SetNextGroup            ; Set params for next group of chunks

        move    #2,playPhase            ; Phase 2 is next

CheckCDExit:
        rts

;------------------------------------------------------------------------------
;   Declare storage for flags and variables.
;------------------------------------------------------------------------------

semaphore:      dc.w    0
CRYmovie:       dc.w    0
dest:           dc.l    0

        .long
timeIncr:       dc.l    0       ; Read by GPU as LONG!

        .long
                dc.w    0       ; Keep this...
time:           dc.l    0       ; +2 is read by GPU as longword!
                dc.w    0

mediaOffset:    dc.l    0
blockOffset:    dc.l    0
cBufBase:       dc.l    0
cBufSize:       dc.l    0
debugHandle:    dc.l    0
GPUOffset:      dc.l    0
deltaTime:      dc.l    0
filmChunks:     dc.l    0
pNextGroup:     dc.l    0
buffChunks:     dc.l    0
nextBufAddr:    dc.l    0
playPhase:      dc.w    0
catchUp:        dc.w    0
maxDelay:       dc.l    0
CDWritePtr:     dc.l    0
subt_begin:     dc.l    0

AudioDesc:      dc.l    0       ; Audio Description Field
MovieWidth:     dc.l    0       ; Width of movie (in pixels)
MovieHeight:    dc.l    0       ; Height of movie (in pixels)
blitScreen:     dc.w    0       ; For double-buffer blit (only high byte is used)

        .long
scrbuf:         dc.l    0       ; Address of buffer to display (Read by GPU as long)
blitCount:      dc.l    0       ; Blit B_COUNT value (precomputed)
blitPixel:      dc.l    0       ; Blit A1/A2_PIXEL value (precomputed)
blitStep:       dc.l    0       ; Blit A1/A2_STEP value (precomputed)

        .even

;------------------------------------------------------------------------------
;   Declare externals and globals.
;------------------------------------------------------------------------------

        .extern DSP_ARGS        ; Where to put DSP arguments
        .extern DSP_ENTR        ; Entry to DSP code
        .extern DSP_STOP        ; DSP run flag?
        .extern AUDIO_DRIFT     ; Audio drift rate
        .extern AUDIO_DESC      ; Audio data description
        .extern GPU_READ        ; GPU has finished startup
        .extern RUN_GPU         ; GPU run flag?

        .extern CheckKeyFrame   ; Check for key frame
        .extern PreDecompress   ; Cinepak codebook expansion
        .extern Decompress      ; Cinepak frame drawing
        .extern HaltCpk         ; Cinepak Shutdown

        .extern Clear           ; Clear display memory
        .extern GPUOLP          ; OLP to be set by GPU

        .extern FindSync        ; Find sync pattern in data stream
        .extern GetCDWritePtr   ; Get location of CD-ROM write
        .extern LoadDSP         ; Load DSP code
        .extern LoadGPU         ; Load GPU code
        .extern LongDivide      ; Long division routine
        .extern ReadCDData      ; Play CD and fill circular buffer
        .extern SetNextGroup    ; Set next chunk group's parameters
        .extern InitROM         ; Clear ROM memory
        .extern Snapshot        ; Make debug snapshot in ROM
        .extern Subtitle        ; Put up subtitle

                                ; Added (SDS)
        .extern MediaGetFirst   ;  |
        .extern MediaGetTrack   ;  v
        .extern MediaGetNext    ;
        .extern IntInit         ;
        .extern DoBlit          ;
        .extern InitBlit        ;

        .extern traknumr
        .extern framenum
        .extern S_frmnum
        .extern E_frmnum
        .extern get_joy

        .globl  MovieWidth      ;  |
        .globl  MovieHeight     ;  v
        .globl  scrbuf          ;
        .globl  blitCount       ;
        .globl  blitPixel       ;
        .globl  blitScreen      ;
        .globl  blitStep        ;  _

        .globl  blockOffset     ; Added (SDS)
        .globl  buffChunks
        .globl  catchUp
        .globl  CDWritePtr
        .globl  debugHandle
        .globl  dest
        .globl  filmChunks
        .globl  GPUOffset
        .globl  mediaOffset
        .globl  playPhase
        .globl  pNextGroup
        .globl  semaphore
        .globl  time
        .globl  timeIncr
        .globl  subt_begin

        .globl  _start          ; Added (SDS)

        .end
